@@ -65,6 +65,10 @@ namespace {
 constexpr size_t max_preferred_versionslen = 4;
 } // namespace
 
+namespace {
+constexpr size_t NGTCP2_STATELESS_RESET_BURST = 100;
+} // namespace
+
 namespace {
 auto randgen = util::make_mt19937();
 } // namespace
@@ -243,9 +247,9 @@ int Stream::start_response() {
   map_file(fe);
 
   if (!config.quiet) {
-    std::array<nghttp3_nv, 1> nva{
+    auto nva = std::to_array({
         util::make_nv_nn(":status", "200"),
-    };
+    });
 
     debug::print_http_response_headers(stream_id, nva.data(), nva.size());
   }
@@ -387,6 +391,9 @@ int Handler::handshake_completed() {
   if (!config.quiet) {
     std::cerr << "Negotiated cipher suite is " << tls_session_.get_cipher_name()
               << std::endl;
+    if (auto group = tls_session_.get_negotiated_group(); !group.empty()) {
+      std::cerr << "Negotiated group is " << group << std::endl;
+    }
     std::cerr << "Negotiated ALPN is " << tls_session_.get_selected_alpn()
               << std::endl;
   }
@@ -432,7 +439,8 @@ int do_hp_mask(uint8_t *dest, const ngtcp2_crypto_cipher *hp,
   }
 
   if (!config.quiet && config.show_secret) {
-    debug::print_hp_mask(dest, NGTCP2_HP_MASKLEN, sample, NGTCP2_HP_SAMPLELEN);
+    debug::print_hp_mask({dest, NGTCP2_HP_MASKLEN},
+                         {sample, NGTCP2_HP_SAMPLELEN});
   }
 
   return 0;
@@ -444,7 +452,7 @@ int recv_crypto_data(ngtcp2_conn *conn,
                      ngtcp2_encryption_level encryption_level, uint64_t offset,
                      const uint8_t *data, size_t datalen, void *user_data) {
   if (!config.quiet && !config.no_quic_dump) {
-    debug::print_crypto_data(encryption_level, data, datalen);
+    debug::print_crypto_data(encryption_level, {data, datalen});
   }
 
   return ngtcp2_crypto_recv_crypto_data_cb(conn, encryption_level, offset, data,
@@ -458,7 +466,7 @@ int recv_stream_data(ngtcp2_conn *conn, uint32_t flags, int64_t stream_id,
                      void *user_data, void *stream_user_data) {
   auto h = static_cast<Handler *>(user_data);
 
-  if (h->recv_stream_data(flags, stream_id, data, datalen) != 0) {
+  if (h->recv_stream_data(flags, stream_id, {data, datalen}) != 0) {
     return NGTCP2_ERR_CALLBACK_FAILURE;
   }
 
@@ -531,7 +539,7 @@ void rand(uint8_t *dest, size_t destlen, const ngtcp2_rand_ctx *rand_ctx) {
 namespace {
 int get_new_connection_id(ngtcp2_conn *conn, ngtcp2_cid *cid, uint8_t *token,
                           size_t cidlen, void *user_data) {
-  if (util::generate_secure_random(cid->data, cidlen) != 0) {
+  if (util::generate_secure_random({cid->data, cidlen}) != 0) {
     return NGTCP2_ERR_CALLBACK_FAILURE;
   }
 
@@ -658,9 +666,8 @@ void Handler::write_qlog(const void *data, size_t datalen) {
 int Handler::init(const Endpoint &ep, const Address &local_addr,
                   const sockaddr *sa, socklen_t salen, const ngtcp2_cid *dcid,
                   const ngtcp2_cid *scid, const ngtcp2_cid *ocid,
-                  const uint8_t *token, size_t tokenlen,
-                  ngtcp2_token_type token_type, uint32_t version,
-                  TLSServerContext &tls_ctx) {
+                  std::span<const uint8_t> token, ngtcp2_token_type token_type,
+                  uint32_t version, TLSServerContext &tls_ctx) {
   auto callbacks = ngtcp2_callbacks{
       nullptr, // client_initial
       ngtcp2_crypto_recv_client_initial_cb,
@@ -704,7 +711,7 @@ int Handler::init(const Endpoint &ep, const Address &local_addr,
   };
 
   scid_.datalen = NGTCP2_SV_SCIDLEN;
-  if (util::generate_secure_random(scid_.data, scid_.datalen) != 0) {
+  if (util::generate_secure_random({scid_.data, scid_.datalen}) != 0) {
     std::cerr << "Could not generate connection ID" << std::endl;
     return -1;
   }
@@ -713,8 +720,8 @@ int Handler::init(const Endpoint &ep, const Address &local_addr,
   ngtcp2_settings_default(&settings);
   settings.log_printf = config.quiet ? nullptr : debug::log_printf;
   settings.initial_ts = util::timestamp();
-  settings.token = token;
-  settings.tokenlen = tokenlen;
+  settings.token = token.data();
+  settings.tokenlen = token.size();
   settings.token_type = token_type;
   settings.cc_algo = config.cc_algo;
   settings.initial_rtt = config.initial_rtt;
@@ -730,7 +737,7 @@ int Handler::init(const Endpoint &ep, const Address &local_addr,
   if (!config.qlog_dir.empty()) {
     auto path = std::string{config.qlog_dir};
     path += '/';
-    path += util::format_hex(scid_.data, scid_.datalen);
+    path += util::format_hex({scid_.data, scid_.datalen});
     path += ".sqlog";
     qlog_ = fopen(path.c_str(), "w");
     if (qlog_ == nullptr) {
@@ -789,9 +796,9 @@ int Handler::init(const Endpoint &ep, const Address &local_addr,
 
   params.original_dcid_present = 1;
 
-  if (util::generate_secure_random(params.stateless_reset_token,
-                                   sizeof(params.stateless_reset_token)) != 0) {
-    std::cerr << "Could not generate stateless reset token" << std::endl;
+  if (ngtcp2_crypto_generate_stateless_reset_token(
+          params.stateless_reset_token, config.static_secret.data(),
+          config.static_secret.size(), &scid_) != 0) {
     return -1;
   }
 
@@ -808,16 +815,17 @@ int Handler::init(const Endpoint &ep, const Address &local_addr,
       params.preferred_addr.ipv6_present = 1;
     }
 
-    auto &token = params.preferred_addr.stateless_reset_token;
-    if (util::generate_secure_random(token, sizeof(token)) != 0) {
+    if (util::generate_secure_random(
+            params.preferred_addr.stateless_reset_token) != 0) {
       std::cerr << "Could not generate preferred address stateless reset token"
                 << std::endl;
       return -1;
     }
 
     params.preferred_addr.cid.datalen = NGTCP2_SV_SCIDLEN;
-    if (util::generate_secure_random(params.preferred_addr.cid.data,
-                                     params.preferred_addr.cid.datalen) != 0) {
+    if (util::generate_secure_random({params.preferred_addr.cid.data,
+                                      params.preferred_addr.cid.datalen}) !=
+        0) {
       std::cerr << "Could not generate preferred address connection ID"
                 << std::endl;
       return -1;
@@ -858,8 +866,8 @@ int Handler::init(const Endpoint &ep, const Address &local_addr,
 
 int Handler::feed_data(const Endpoint &ep, const Address &local_addr,
                        const sockaddr *sa, socklen_t salen,
-                       const ngtcp2_pkt_info *pi, const uint8_t *data,
-                       size_t datalen) {
+                       const ngtcp2_pkt_info *pi,
+                       std::span<const uint8_t> data) {
   auto path = ngtcp2_path{
       {
           const_cast<sockaddr *>(&local_addr.su.sa),
@@ -872,7 +880,7 @@ int Handler::feed_data(const Endpoint &ep, const Address &local_addr,
       const_cast<Endpoint *>(&ep),
   };
 
-  if (auto rv = ngtcp2_conn_read_pkt(conn_, &path, pi, data, datalen,
+  if (auto rv = ngtcp2_conn_read_pkt(conn_, &path, pi, data.data(), data.size(),
                                      util::timestamp());
       rv != 0) {
     std::cerr << "ngtcp2_conn_read_pkt: " << ngtcp2_strerror(rv) << std::endl;
@@ -903,10 +911,8 @@ int Handler::feed_data(const Endpoint &ep, const Address &local_addr,
 
 int Handler::on_read(const Endpoint &ep, const Address &local_addr,
                      const sockaddr *sa, socklen_t salen,
-                     const ngtcp2_pkt_info *pi, const uint8_t *data,
-                     size_t datalen) {
-  if (auto rv = feed_data(ep, local_addr, sa, salen, pi, data, datalen);
-      rv != 0) {
+                     const ngtcp2_pkt_info *pi, std::span<const uint8_t> data) {
+  if (auto rv = feed_data(ep, local_addr, sa, salen, pi, data); rv != 0) {
     return rv;
   }
 
@@ -958,23 +964,20 @@ int Handler::write_streams() {
   ngtcp2_vec vec;
   ngtcp2_path_storage ps, prev_ps;
   uint32_t prev_ecn = 0;
-  size_t pktcnt = 0;
   auto max_udp_payload_size = ngtcp2_conn_get_max_tx_udp_payload_size(conn_);
   auto path_max_udp_payload_size =
       ngtcp2_conn_get_path_max_tx_udp_payload_size(conn_);
-  auto max_pktcnt =
-      std::max(ngtcp2_conn_get_send_quantum(conn_) / max_udp_payload_size,
-               static_cast<size_t>(1));
-  uint8_t *bufpos = tx_.data.get();
   ngtcp2_pkt_info pi;
   size_t gso_size = 0;
   auto ts = util::timestamp();
+  auto txbuf =
+      std::span{tx_.data.get(), std::max(ngtcp2_conn_get_send_quantum(conn_),
+                                         path_max_udp_payload_size)};
+  auto buf = txbuf;
 
   ngtcp2_path_storage_zero(&ps);
   ngtcp2_path_storage_zero(&prev_ps);
 
-  max_pktcnt = std::min(max_pktcnt, static_cast<size_t>(config.max_gso_dgrams));
-
   for (;;) {
     int64_t stream_id = -1;
     size_t vcnt = 0;
@@ -993,9 +996,12 @@ int Handler::write_streams() {
 
     ngtcp2_ssize ndatalen;
 
-    auto nwrite = ngtcp2_conn_writev_stream(conn_, &ps.path, &pi, bufpos,
-                                            max_udp_payload_size, &ndatalen,
-                                            flags, stream_id, &vec, vcnt, ts);
+    auto buflen = buf.size() >= max_udp_payload_size
+                      ? max_udp_payload_size
+                      : path_max_udp_payload_size;
+    auto nwrite =
+        ngtcp2_conn_writev_stream(conn_, &ps.path, &pi, buf.data(), buflen,
+                                  &ndatalen, flags, stream_id, &vec, vcnt, ts);
     if (nwrite < 0) {
       switch (nwrite) {
       case NGTCP2_ERR_STREAM_DATA_BLOCKED:
@@ -1026,19 +1032,18 @@ int Handler::write_streams() {
     }
 
     if (nwrite == 0) {
-      if (bufpos - tx_.data.get()) {
+      auto data = std::span{std::begin(txbuf), std::begin(buf)};
+      if (!data.empty()) {
         auto &ep = *static_cast<Endpoint *>(prev_ps.path.user_data);
-        auto data = tx_.data.get();
-        auto datalen = bufpos - data;
 
-        if (auto [nsent, rv] = server_->send_packet(
+        if (auto [rest, rv] = server_->send_packet(
                 ep, no_gso_, prev_ps.path.local, prev_ps.path.remote, prev_ecn,
-                data, datalen, gso_size);
+                data, gso_size);
             rv != NETWORK_ERR_OK) {
           assert(NETWORK_ERR_SEND_BLOCKED == rv);
 
           on_send_blocked(ep, prev_ps.path.local, prev_ps.path.remote, prev_ecn,
-                          data + nsent, datalen - nsent, gso_size);
+                          rest, gso_size);
 
           start_wev_endpoint(ep);
         }
@@ -1049,9 +1054,11 @@ int Handler::write_streams() {
       return 0;
     }
 
-    bufpos += nwrite;
+    auto last_pkt = std::begin(buf);
 
-    if (pktcnt == 0) {
+    buf = buf.subspan(nwrite);
+
+    if (last_pkt == std::begin(txbuf)) {
       ngtcp2_path_copy(&prev_ps.path, &ps.path);
       prev_ecn = pi.ecn;
       gso_size = nwrite;
@@ -1060,36 +1067,36 @@ int Handler::write_streams() {
                (gso_size > path_max_udp_payload_size &&
                 static_cast<size_t>(nwrite) != gso_size)) {
       auto &ep = *static_cast<Endpoint *>(prev_ps.path.user_data);
-      auto data = tx_.data.get();
-      auto datalen = bufpos - data - nwrite;
+      auto data = std::span{std::begin(txbuf), last_pkt};
 
-      if (auto [nsent, rv] = server_->send_packet(
+      if (auto [rest, rv] = server_->send_packet(
               ep, no_gso_, prev_ps.path.local, prev_ps.path.remote, prev_ecn,
-              data, datalen, gso_size);
+              data, gso_size);
           rv != 0) {
         assert(NETWORK_ERR_SEND_BLOCKED == rv);
 
         on_send_blocked(ep, prev_ps.path.local, prev_ps.path.remote, prev_ecn,
-                        data + nsent, datalen - nsent, gso_size);
+                        rest, gso_size);
 
+        data = std::span{last_pkt, std::begin(buf)};
         on_send_blocked(*static_cast<Endpoint *>(ps.path.user_data),
-                        ps.path.local, ps.path.remote, pi.ecn, bufpos - nwrite,
-                        nwrite, 0);
+                        ps.path.local, ps.path.remote, pi.ecn, data,
+                        data.size());
 
         start_wev_endpoint(ep);
       } else {
         auto &ep = *static_cast<Endpoint *>(ps.path.user_data);
-        auto data = bufpos - nwrite;
+        auto data = std::span{last_pkt, std::begin(buf)};
 
-        if (auto [nsent, rv] =
+        if (auto [rest, rv] =
                 server_->send_packet(ep, no_gso_, ps.path.local, ps.path.remote,
-                                     pi.ecn, data, nwrite, nwrite);
+                                     pi.ecn, data, data.size());
             rv != 0) {
-          assert(nsent == 0);
+          assert(rest.size() == data.size());
           assert(NETWORK_ERR_SEND_BLOCKED == rv);
 
-          on_send_blocked(ep, ps.path.local, ps.path.remote, pi.ecn, data,
-                          nwrite, 0);
+          on_send_blocked(ep, ps.path.local, ps.path.remote, pi.ecn, rest,
+                          rest.size());
 
           start_wev_endpoint(ep);
         }
@@ -1099,19 +1106,19 @@ int Handler::write_streams() {
       return 0;
     }
 
-    if (++pktcnt == max_pktcnt || static_cast<size_t>(nwrite) < gso_size) {
+    if (buf.size() < path_max_udp_payload_size ||
+        static_cast<size_t>(nwrite) < gso_size) {
       auto &ep = *static_cast<Endpoint *>(ps.path.user_data);
-      auto data = tx_.data.get();
-      auto datalen = bufpos - data;
+      auto data = std::span{std::begin(txbuf), std::begin(buf)};
 
-      if (auto [nsent, rv] =
+      if (auto [rest, rv] =
               server_->send_packet(ep, no_gso_, ps.path.local, ps.path.remote,
-                                   pi.ecn, data, datalen, gso_size);
+                                   pi.ecn, data, gso_size);
           rv != 0) {
         assert(NETWORK_ERR_SEND_BLOCKED == rv);
 
-        on_send_blocked(ep, ps.path.local, ps.path.remote, pi.ecn, data + nsent,
-                        datalen - nsent, gso_size);
+        on_send_blocked(ep, ps.path.local, ps.path.remote, pi.ecn, rest,
+                        gso_size);
 
         start_wev_endpoint(ep);
       }
@@ -1124,10 +1131,10 @@ int Handler::write_streams() {
 
 void Handler::on_send_blocked(Endpoint &ep, const ngtcp2_addr &local_addr,
                               const ngtcp2_addr &remote_addr, unsigned int ecn,
-                              const uint8_t *data, size_t datalen,
-                              size_t gso_size) {
+                              std::span<const uint8_t> data, size_t gso_size) {
   assert(tx_.num_blocked || !tx_.send_blocked);
   assert(tx_.num_blocked < 2);
+  assert(gso_size);
 
   tx_.send_blocked = true;
 
@@ -1141,7 +1148,6 @@ void Handler::on_send_blocked(Endpoint &ep, const ngtcp2_addr &local_addr,
   p.endpoint = &ep;
   p.ecn = ecn;
   p.data = data;
-  p.datalen = datalen;
   p.gso_size = gso_size;
 }
 
@@ -1174,14 +1180,13 @@ int Handler::send_blocked_packet() {
         .addrlen = p.remote_addr.len,
     };
 
-    auto [nsent, rv] =
+    auto [rest, rv] =
         server_->send_packet(*p.endpoint, no_gso_, local_addr, remote_addr,
-                             p.ecn, p.data, p.datalen, p.gso_size);
+                             p.ecn, p.data, p.gso_size);
     if (rv != 0) {
       assert(NETWORK_ERR_SEND_BLOCKED == rv);
 
-      p.data += nsent;
-      p.datalen -= nsent;
+      p.data = rest;
 
       start_wev_endpoint(*p.endpoint);
 
@@ -1286,9 +1291,9 @@ int Handler::send_conn_close() {
 
   auto path = ngtcp2_conn_get_path(conn_);
 
-  return server_->send_packet(
-      *static_cast<Endpoint *>(path->user_data), path->local, path->remote,
-      /* ecn = */ 0, conn_closebuf_->rpos(), conn_closebuf_->size());
+  return server_->send_packet(*static_cast<Endpoint *>(path->user_data),
+                              path->local, path->remote,
+                              /* ecn = */ 0, conn_closebuf_->data());
 }
 
 void Handler::update_timer() {
@@ -1359,9 +1364,9 @@ auto htp_settings = http_parser_settings{
 };
 
 int Handler::recv_stream_data(uint32_t flags, int64_t stream_id,
-                              const uint8_t *data, size_t datalen) {
+                              std::span<const uint8_t> data) {
   if (!config.quiet && !config.no_quic_dump) {
-    debug::print_stream_data(stream_id, data, datalen);
+    debug::print_stream_data(stream_id, data);
   }
 
   auto it = streams_.find(stream_id);
@@ -1369,10 +1374,10 @@ int Handler::recv_stream_data(uint32_t flags, int64_t stream_id,
   auto &stream = (*it).second;
 
   if (!stream->eos) {
-    auto nread =
-        http_parser_execute(&stream->htp, &htp_settings,
-                            reinterpret_cast<const char *>(data), datalen);
-    if (nread != datalen) {
+    auto nread = http_parser_execute(
+        &stream->htp, &htp_settings,
+        reinterpret_cast<const char *>(data.data()), data.size());
+    if (nread != data.size()) {
       if (auto rv = ngtcp2_conn_shutdown_stream(conn_, 0, stream_id,
                                                 /* app error code */ 1);
           rv != 0) {
@@ -1384,8 +1389,8 @@ int Handler::recv_stream_data(uint32_t flags, int64_t stream_id,
     }
   }
 
-  ngtcp2_conn_extend_max_stream_offset(conn_, stream_id, datalen);
-  ngtcp2_conn_extend_max_offset(conn_, datalen);
+  ngtcp2_conn_extend_max_stream_offset(conn_, stream_id, data.size());
+  ngtcp2_conn_extend_max_offset(conn_, data.size());
 
   return 0;
 }
@@ -1413,11 +1418,11 @@ int Handler::update_key(uint8_t *rx_secret, uint8_t *tx_secret,
 
   if (!config.quiet && config.show_secret) {
     std::cerr << "application_traffic rx secret " << nkey_update_ << std::endl;
-    debug::print_secrets(rx_secret, secretlen, rx_key.data(), keylen, rx_iv,
-                         ivlen);
+    debug::print_secrets({rx_secret, secretlen}, {rx_key.data(), keylen},
+                         {rx_iv, ivlen});
     std::cerr << "application_traffic tx secret " << nkey_update_ << std::endl;
-    debug::print_secrets(tx_secret, secretlen, tx_key.data(), keylen, tx_iv,
-                         ivlen);
+    debug::print_secrets({tx_secret, secretlen}, {tx_key.data(), keylen},
+                         {tx_iv, ivlen});
   }
 
   return 0;
@@ -1472,8 +1477,20 @@ void siginthandler(struct ev_loop *loop, ev_signal *watcher, int revents) {
 } // namespace
 
 Server::Server(struct ev_loop *loop, TLSServerContext &tls_ctx)
-    : loop_(loop), tls_ctx_(tls_ctx) {
+    : loop_(loop),
+      tls_ctx_(tls_ctx),
+      stateless_reset_bucket_(NGTCP2_STATELESS_RESET_BURST) {
   ev_signal_init(&sigintev_, siginthandler, SIGINT);
+
+  ev_timer_init(
+      &stateless_reset_regen_timer_,
+      [](struct ev_loop *loop, ev_timer *w, int revents) {
+        auto server = static_cast<Server *>(w->data);
+
+        server->on_stateless_reset_regen();
+      },
+      0., 1.);
+  stateless_reset_regen_timer_.data = this;
 }
 
 Server::~Server() {
@@ -1488,6 +1505,7 @@ void Server::disconnect() {
     ev_io_stop(loop_, &ep.rev);
   }
 
+  ev_timer_stop(loop_, &stateless_reset_regen_timer_);
   ev_signal_stop(loop_, &sigintev_);
 
   while (!handlers_.empty()) {
@@ -1738,6 +1756,19 @@ int Server::on_read(Endpoint &ep) {
       return 0;
     }
 
+    // Packets less than 22 bytes never be a valid QUIC packet.
+    if (nread < 22) {
+      ++pktcnt;
+
+      continue;
+    }
+
+    if (util::prohibited_port(util::port(&su))) {
+      ++pktcnt;
+
+      continue;
+    }
+
     pi.ecn = msghdr_get_ecn(&msg, su.storage.ss_family);
     auto local_addr = msghdr_get_local_addr(&msg, su.storage.ss_family);
     if (!local_addr) {
@@ -1753,10 +1784,10 @@ int Server::on_read(Endpoint &ep) {
 
     set_port(*local_addr, ep.addr);
 
-    auto data = buf.data();
+    auto data = std::span{buf.data(), static_cast<size_t>(nread)};
 
-    for (;;) {
-      auto datalen = std::min(static_cast<size_t>(nread), gso_size);
+    for (; !data.empty();) {
+      auto datalen = std::min(data.size(), gso_size);
 
       ++pktcnt;
 
@@ -1771,20 +1802,21 @@ int Server::on_read(Endpoint &ep) {
                   << std::dec << " " << datalen << " bytes" << std::endl;
       }
 
+      // Packets less than 22 bytes never be a valid QUIC packet.
+      if (datalen < 22) {
+        break;
+      }
+
       if (debug::packet_lost(config.rx_loss_prob)) {
         if (!config.quiet) {
           std::cerr << "** Simulated incoming packet loss **" << std::endl;
         }
       } else {
-        read_pkt(ep, *local_addr, &su.sa, msg.msg_namelen, &pi, data, datalen);
+        read_pkt(ep, *local_addr, &su.sa, msg.msg_namelen, &pi,
+                 {data.data(), datalen});
       }
 
-      nread -= datalen;
-      if (nread == 0) {
-        break;
-      }
-
-      data += datalen;
+      data = data.subspan(datalen);
     }
   }
 
@@ -1793,22 +1825,18 @@ int Server::on_read(Endpoint &ep) {
 
 void Server::read_pkt(Endpoint &ep, const Address &local_addr,
                       const sockaddr *sa, socklen_t salen,
-                      const ngtcp2_pkt_info *pi, const uint8_t *data,
-                      size_t datalen) {
-  if (datalen == 0) {
-    return;
-  }
-
+                      const ngtcp2_pkt_info *pi,
+                      std::span<const uint8_t> data) {
   ngtcp2_version_cid vc;
 
-  switch (auto rv = ngtcp2_pkt_decode_version_cid(&vc, data, datalen,
+  switch (auto rv = ngtcp2_pkt_decode_version_cid(&vc, data.data(), data.size(),
                                                   NGTCP2_SV_SCIDLEN);
           rv) {
   case 0:
     break;
   case NGTCP2_ERR_VERSION_NEGOTIATION:
-    send_version_negotiation(vc.version, vc.scid, vc.scidlen, vc.dcid,
-                             vc.dcidlen, ep, local_addr, sa, salen);
+    send_version_negotiation(vc.version, {vc.scid, vc.scidlen},
+                             {vc.dcid, vc.dcidlen}, ep, local_addr, sa, salen);
     return;
   default:
     std::cerr << "Could not decode version and CID from QUIC packet header: "
@@ -1816,17 +1844,23 @@ void Server::read_pkt(Endpoint &ep, const Address &local_addr,
     return;
   }
 
-  auto dcid_key = util::make_cid_key(vc.dcid, vc.dcidlen);
+  auto dcid_key = util::make_cid_key({vc.dcid, vc.dcidlen});
 
   auto handler_it = handlers_.find(dcid_key);
   if (handler_it == std::end(handlers_)) {
     ngtcp2_pkt_hd hd;
 
-    if (auto rv = ngtcp2_accept(&hd, data, datalen); rv != 0) {
+    if (auto rv = ngtcp2_accept(&hd, data.data(), data.size()); rv != 0) {
       if (!config.quiet) {
-        std::cerr << "Unexpected packet received: length=" << datalen
+        std::cerr << "Unexpected packet received: length=" << data.size()
                   << std::endl;
       }
+
+      if (!(data[0] & 0x80) && data.size() >= NGTCP2_SV_SCIDLEN + 22) {
+        send_stateless_reset(data.size(), {vc.dcid, vc.dcidlen}, ep, local_addr,
+                             sa, salen);
+      }
+
       return;
     }
 
@@ -1839,7 +1873,7 @@ void Server::read_pkt(Endpoint &ep, const Address &local_addr,
     if (config.validate_addr || hd.tokenlen) {
       std::cerr << "Perform stateless address validation" << std::endl;
       if (hd.tokenlen == 0) {
-        send_retry(&hd, ep, local_addr, sa, salen, datalen * 3);
+        send_retry(&hd, ep, local_addr, sa, salen, data.size() * 3);
         return;
       }
 
@@ -1863,7 +1897,7 @@ void Server::read_pkt(Endpoint &ep, const Address &local_addr,
       case NGTCP2_CRYPTO_TOKEN_MAGIC_REGULAR:
         if (verify_token(&hd, sa, salen) != 0) {
           if (config.validate_addr) {
-            send_retry(&hd, ep, local_addr, sa, salen, datalen * 3);
+            send_retry(&hd, ep, local_addr, sa, salen, data.size() * 3);
             return;
           }
 
@@ -1878,7 +1912,7 @@ void Server::read_pkt(Endpoint &ep, const Address &local_addr,
           std::cerr << "Ignore unrecognized token" << std::endl;
         }
         if (config.validate_addr) {
-          send_retry(&hd, ep, local_addr, sa, salen, datalen * 3);
+          send_retry(&hd, ep, local_addr, sa, salen, data.size() * 3);
           return;
         }
 
@@ -1889,16 +1923,17 @@ void Server::read_pkt(Endpoint &ep, const Address &local_addr,
     }
 
     auto h = std::make_unique<Handler>(loop_, this);
-    if (h->init(ep, local_addr, sa, salen, &hd.scid, &hd.dcid, pocid, hd.token,
-                hd.tokenlen, token_type, hd.version, tls_ctx_) != 0) {
+    if (h->init(ep, local_addr, sa, salen, &hd.scid, &hd.dcid, pocid,
+                {hd.token, hd.tokenlen}, token_type, hd.version,
+                tls_ctx_) != 0) {
       return;
     }
 
-    switch (h->on_read(ep, local_addr, sa, salen, pi, data, datalen)) {
+    switch (h->on_read(ep, local_addr, sa, salen, pi, data)) {
     case 0:
       break;
     case NETWORK_ERR_RETRY:
-      send_retry(&hd, ep, local_addr, sa, salen, datalen * 3);
+      send_retry(&hd, ep, local_addr, sa, salen, data.size() * 3);
       return;
     default:
       return;
@@ -1939,8 +1974,7 @@ void Server::read_pkt(Endpoint &ep, const Address &local_addr,
     return;
   }
 
-  if (auto rv = h->on_read(ep, local_addr, sa, salen, pi, data, datalen);
-      rv != 0) {
+  if (auto rv = h->on_read(ep, local_addr, sa, salen, pi, data); rv != 0) {
     if (rv != NETWORK_ERR_CLOSE_WAIT) {
       remove(h);
     }
@@ -1973,10 +2007,10 @@ uint32_t generate_reserved_version(const sockaddr *sa, socklen_t salen,
 }
 } // namespace
 
-int Server::send_version_negotiation(uint32_t version, const uint8_t *dcid,
-                                     size_t dcidlen, const uint8_t *scid,
-                                     size_t scidlen, Endpoint &ep,
-                                     const Address &local_addr,
+int Server::send_version_negotiation(uint32_t version,
+                                     std::span<const uint8_t> dcid,
+                                     std::span<const uint8_t> scid,
+                                     Endpoint &ep, const Address &local_addr,
                                      const sockaddr *sa, socklen_t salen) {
   Buffer buf{NGTCP2_MAX_UDP_PAYLOAD_SIZE};
   std::array<uint32_t, 1 + max_preferred_versionslen> sv;
@@ -1995,7 +2029,8 @@ int Server::send_version_negotiation(uint32_t version, const uint8_t *dcid,
 
   auto nwrite = ngtcp2_pkt_write_version_negotiation(
       buf.wpos(), buf.left(), std::uniform_int_distribution<uint8_t>()(randgen),
-      dcid, dcidlen, scid, scidlen, sv.data(), p - std::begin(sv));
+      dcid.data(), dcid.size(), scid.data(), scid.size(), sv.data(),
+      p - std::begin(sv));
   if (nwrite < 0) {
     std::cerr << "ngtcp2_pkt_write_version_negotiation: "
               << ngtcp2_strerror(nwrite) << std::endl;
@@ -2013,7 +2048,7 @@ int Server::send_version_negotiation(uint32_t version, const uint8_t *dcid,
       salen,
   };
 
-  if (send_packet(ep, laddr, raddr, /* ecn = */ 0, buf.rpos(), buf.size()) !=
+  if (send_packet(ep, laddr, raddr, /* ecn = */ 0, buf.data()) !=
       NETWORK_ERR_OK) {
     return -1;
   }
@@ -2042,7 +2077,7 @@ int Server::send_retry(const ngtcp2_pkt_hd *chd, Endpoint &ep,
   ngtcp2_cid scid;
 
   scid.datalen = NGTCP2_SV_SCIDLEN;
-  if (util::generate_secure_random(scid.data, scid.datalen) != 0) {
+  if (util::generate_secure_random({scid.data, scid.datalen}) != 0) {
     return -1;
   }
 
@@ -2086,7 +2121,7 @@ int Server::send_retry(const ngtcp2_pkt_hd *chd, Endpoint &ep,
       salen,
   };
 
-  if (send_packet(ep, laddr, raddr, /* ecn = */ 0, buf.rpos(), buf.size()) !=
+  if (send_packet(ep, laddr, raddr, /* ecn = */ 0, buf.data()) !=
       NETWORK_ERR_OK) {
     return -1;
   }
@@ -2120,7 +2155,82 @@ int Server::send_stateless_connection_close(const ngtcp2_pkt_hd *chd,
       salen,
   };
 
-  if (send_packet(ep, laddr, raddr, /* ecn = */ 0, buf.rpos(), buf.size()) !=
+  if (send_packet(ep, laddr, raddr, /* ecn = */ 0, buf.data()) !=
+      NETWORK_ERR_OK) {
+    return -1;
+  }
+
+  return 0;
+}
+
+int Server::send_stateless_reset(size_t pktlen, std::span<const uint8_t> dcid,
+                                 Endpoint &ep, const Address &local_addr,
+                                 const sockaddr *sa, socklen_t salen) {
+  if (stateless_reset_bucket_ == 0) {
+    return 0;
+  }
+
+  --stateless_reset_bucket_;
+
+  if (!ev_is_active(&stateless_reset_regen_timer_)) {
+    ev_timer_again(loop_, &stateless_reset_regen_timer_);
+  }
+
+  ngtcp2_cid cid;
+
+  ngtcp2_cid_init(&cid, dcid.data(), dcid.size());
+
+  std::array<uint8_t, NGTCP2_STATELESS_RESET_TOKENLEN> token;
+
+  if (ngtcp2_crypto_generate_stateless_reset_token(
+          token.data(), config.static_secret.data(),
+          config.static_secret.size(), &cid) != 0) {
+    return -1;
+  }
+
+  // SCID + minimum expansion - NGTCP2_STATELESS_RESET_TOKENLEN
+  constexpr size_t max_rand_byteslen =
+      NGTCP2_MAX_CIDLEN + 22 - NGTCP2_STATELESS_RESET_TOKENLEN;
+
+  size_t rand_byteslen;
+
+  if (pktlen <= 43) {
+    // As per
+    // https://datatracker.ietf.org/doc/html/rfc9000#section-10.3
+    rand_byteslen = pktlen - NGTCP2_STATELESS_RESET_TOKENLEN - 1;
+  } else {
+    rand_byteslen = max_rand_byteslen;
+  }
+
+  std::array<uint8_t, max_rand_byteslen> rand_bytes;
+
+  if (util::generate_secure_random({rand_bytes.data(), rand_byteslen}) != 0) {
+    return -1;
+  }
+
+  Buffer buf{NGTCP2_MAX_UDP_PAYLOAD_SIZE};
+
+  auto nwrite = ngtcp2_pkt_write_stateless_reset(
+      buf.wpos(), buf.left(), token.data(), rand_bytes.data(), rand_byteslen);
+  if (nwrite < 0) {
+    std::cerr << "ngtcp2_pkt_write_stateless_reset: " << ngtcp2_strerror(nwrite)
+              << std::endl;
+
+    return -1;
+  }
+
+  buf.push(nwrite);
+
+  ngtcp2_addr laddr{
+      const_cast<sockaddr *>(&local_addr.su.sa),
+      local_addr.len,
+  };
+  ngtcp2_addr raddr{
+      const_cast<sockaddr *>(sa),
+      salen,
+  };
+
+  if (send_packet(ep, laddr, raddr, /* ecn = */ 0, buf.data()) !=
       NETWORK_ERR_OK) {
     return -1;
   }
@@ -2206,48 +2316,46 @@ int Server::verify_token(const ngtcp2_pkt_hd *hd, const sockaddr *sa,
 
 int Server::send_packet(Endpoint &ep, const ngtcp2_addr &local_addr,
                         const ngtcp2_addr &remote_addr, unsigned int ecn,
-                        const uint8_t *data, size_t datalen) {
+                        std::span<const uint8_t> data) {
   auto no_gso = false;
-  auto [_, rv] = send_packet(ep, no_gso, local_addr, remote_addr, ecn, data,
-                             datalen, datalen);
+  auto [_, rv] =
+      send_packet(ep, no_gso, local_addr, remote_addr, ecn, data, data.size());
 
   return rv;
 }
 
-std::pair<size_t, int>
+std::pair<std::span<const uint8_t>, int>
 Server::send_packet(Endpoint &ep, bool &no_gso, const ngtcp2_addr &local_addr,
                     const ngtcp2_addr &remote_addr, unsigned int ecn,
-                    const uint8_t *data, size_t datalen, size_t gso_size) {
+                    std::span<const uint8_t> data, size_t gso_size) {
   assert(gso_size);
 
   if (debug::packet_lost(config.tx_loss_prob)) {
     if (!config.quiet) {
       std::cerr << "** Simulated outgoing packet loss **" << std::endl;
     }
-    return {0, NETWORK_ERR_OK};
+    return {{}, NETWORK_ERR_OK};
   }
 
-  if (no_gso && datalen > gso_size) {
-    size_t nsent = 0;
+  if (no_gso && data.size() > gso_size) {
+    for (; !data.empty();) {
+      auto len = std::min(gso_size, data.size());
 
-    for (auto p = data; p < data + datalen; p += gso_size) {
-      auto len = std::min(gso_size, static_cast<size_t>(data + datalen - p));
-
-      auto [n, rv] =
-          send_packet(ep, no_gso, local_addr, remote_addr, ecn, p, len, len);
+      auto [_, rv] = send_packet(ep, no_gso, local_addr, remote_addr, ecn,
+                                 {std::begin(data), len}, len);
       if (rv != 0) {
-        return {nsent, rv};
+        return {data, rv};
       }
 
-      nsent += n;
+      data = data.subspan(len);
     }
 
-    return {nsent, 0};
+    return {{}, 0};
   }
 
   iovec msg_iov;
-  msg_iov.iov_base = const_cast<uint8_t *>(data);
-  msg_iov.iov_len = datalen;
+  msg_iov.iov_base = const_cast<uint8_t *>(data.data());
+  msg_iov.iov_len = data.size();
 
   msghdr msg{};
   msg.msg_name = const_cast<sockaddr *>(remote_addr.addr);
@@ -2297,7 +2405,7 @@ Server::send_packet(Endpoint &ep, bool &no_gso, const ngtcp2_addr &local_addr,
   }
 
 #ifdef UDP_SEGMENT
-  if (datalen > gso_size) {
+  if (data.size() > gso_size) {
     controllen += CMSG_SPACE(sizeof(uint16_t));
     cm = CMSG_NXTHDR(&msg, cm);
     cm->cmsg_level = SOL_UDP;
@@ -2342,10 +2450,10 @@ Server::send_packet(Endpoint &ep, bool &no_gso, const ngtcp2_addr &local_addr,
 #if EAGAIN != EWOULDBLOCK
     case EWOULDBLOCK:
 #endif // EAGAIN != EWOULDBLOCK
-      return {0, NETWORK_ERR_SEND_BLOCKED};
+      return {data, NETWORK_ERR_SEND_BLOCKED};
 #ifdef UDP_SEGMENT
     case EIO:
-      if (datalen > gso_size) {
+      if (data.size() > gso_size) {
         // GSO failure; send each packet in a separate sendmsg call.
         std::cerr << "sendmsg: disabling GSO due to " << strerror(errno)
                   << std::endl;
@@ -2353,7 +2461,7 @@ Server::send_packet(Endpoint &ep, bool &no_gso, const ngtcp2_addr &local_addr,
         no_gso = true;
 
         return send_packet(ep, no_gso, local_addr, remote_addr, ecn, data,
-                           datalen, gso_size);
+                           gso_size);
       }
       break;
 #endif // UDP_SEGMENT
@@ -2362,7 +2470,7 @@ Server::send_packet(Endpoint &ep, bool &no_gso, const ngtcp2_addr &local_addr,
     std::cerr << "sendmsg: " << strerror(errno) << std::endl;
     // TODO We have packet which is expected to fail to send (e.g.,
     // path validation to old path).
-    return {0, NETWORK_ERR_OK};
+    return {{}, NETWORK_ERR_OK};
   }
 
   if (!config.quiet) {
@@ -2374,7 +2482,7 @@ Server::send_packet(Endpoint &ep, bool &no_gso, const ngtcp2_addr &local_addr,
               << " bytes" << std::endl;
   }
 
-  return {nwrite, NETWORK_ERR_OK};
+  return {{}, NETWORK_ERR_OK};
 }
 
 void Server::associate_cid(const ngtcp2_cid *cid, Handler *h) {
@@ -2382,7 +2490,7 @@ void Server::associate_cid(const ngtcp2_cid *cid, Handler *h) {
 }
 
 void Server::dissociate_cid(const ngtcp2_cid *cid) {
-  handlers_.erase(util::make_cid_key(cid));
+  handlers_.erase(std::string{util::make_cid_key(cid)});
 }
 
 void Server::remove(const Handler *h) {
@@ -2400,6 +2508,14 @@ void Server::remove(const Handler *h) {
   delete h;
 }
 
+void Server::on_stateless_reset_regen() {
+  assert(stateless_reset_bucket_ < NGTCP2_STATELESS_RESET_BURST);
+
+  if (++stateless_reset_bucket_ == NGTCP2_STATELESS_RESET_BURST) {
+    ev_timer_stop(loop_, &stateless_reset_regen_timer_);
+  }
+}
+
 namespace {
 int parse_host_port(Address &dest, int af, const char *first,
                     const char *last) {
@@ -2493,7 +2609,6 @@ void config_set_default(Config &config) {
   config.max_dyn_length = 20_m;
   config.cc_algo = NGTCP2_CC_ALGO_CUBIC;
   config.initial_rtt = NGTCP2_DEFAULT_INITIAL_RTT;
-  config.max_gso_dgrams = 64;
   config.handshake_timeout = UINT64_MAX;
   config.ack_thresh = 2;
   config.initial_pkt_num = UINT32_MAX;
@@ -2629,11 +2744,6 @@ Options:
             << util::format_uint_iec(config.max_stream_window) << R"(
   --send-trailers
               Send trailer fields.
-  --max-gso-dgrams=<N>
-              Maximum  number of  UDP  datagrams that  are  sent in  a
-              single GSO sendmsg call.
-              Default: )"
-            << config.max_gso_dgrams << R"(
   --handshake-timeout=<DURATION>
               Set  the  QUIC handshake  timeout.   It  defaults to  no
               timeout.
@@ -2729,7 +2839,6 @@ int main(int argc, char **argv) {
         {"send-trailers", no_argument, &flag, 22},
         {"max-window", required_argument, &flag, 23},
         {"max-stream-window", required_argument, &flag, 24},
-        {"max-gso-dgrams", required_argument, &flag, 25},
         {"handshake-timeout", required_argument, &flag, 26},
         {"preferred-versions", required_argument, &flag, 27},
         {"available-versions", required_argument, &flag, 28},
@@ -2969,15 +3078,6 @@ int main(int argc, char **argv) {
           config.max_stream_window = *n;
         }
         break;
-      case 25:
-        // --max-gso-dgrams
-        if (auto n = util::parse_uint(optarg); !n) {
-          std::cerr << "max-gso-dgrams: invalid argument" << std::endl;
-          exit(EXIT_FAILURE);
-        } else {
-          config.max_gso_dgrams = *n;
-        }
-        break;
       case 26:
         // --handshake-timeout
         if (auto t = util::parse_duration(optarg); !t) {
@@ -3147,8 +3247,7 @@ int main(int argc, char **argv) {
     }
   }
 
-  if (util::generate_secret(config.static_secret.data(),
-                            config.static_secret.size()) != 0) {
+  if (util::generate_secret(config.static_secret) != 0) {
     std::cerr << "Unable to generate static secret" << std::endl;
     exit(EXIT_FAILURE);
   }
